@@ -7,12 +7,12 @@
 use crate::block::hashtable::get_table_size;
 use crate::block::hashtable::HashTable;
 use crate::block::hashtable::{HashTableU16, HashTableU32, HashTableUsize};
-use crate::block::Sink;
 use crate::block::END_OFFSET;
 use crate::block::LZ4_MIN_LENGTH;
 use crate::block::MAX_DISTANCE;
 use crate::block::MFLIMIT;
 use crate::block::MINMATCH;
+use crate::sink::{vec_sink_for_compression, Sink, SliceSink};
 use alloc::vec::Vec;
 
 #[cfg(feature = "safe-encode")]
@@ -202,7 +202,7 @@ fn count_same_bytes(input: &[u8], cur: &mut usize, source: &[u8], candidate: usi
         }
     }
 
-    if *cur + 1 <= input_end
+    if *cur < input_end
         && unsafe { input.as_ptr().add(*cur).read() } == unsafe { source_ptr.read() }
     {
         *cur += 1;
@@ -214,7 +214,7 @@ fn count_same_bytes(input: &[u8], cur: &mut usize, source: &[u8], candidate: usi
 // Write an integer to the output in LSIC format.
 #[inline]
 #[cfg(feature = "safe-encode")]
-fn write_integer(output: &mut Sink, mut n: usize) {
+fn write_integer(output: &mut impl Sink, mut n: usize) {
     // Note: Since `n` is usually < 0xFF and writing multiple bytes to the output
     // requires 2 branches of bound check (due to the possibility of add overflows)
     // the simple byte at a time implementation bellow is faster in most cases.
@@ -228,7 +228,7 @@ fn write_integer(output: &mut Sink, mut n: usize) {
 /// Write an integer to the output in LSIC format.
 #[inline]
 #[cfg(not(feature = "safe-encode"))]
-fn write_integer(output: &mut Sink, mut n: usize) {
+fn write_integer(output: &mut impl Sink, mut n: usize) {
     // Write the 0xFF bytes as long as the integer is higher than said value.
     if n >= 4 * 0xFF {
         // In this unlikelly branch we use a fill instead of a loop,
@@ -236,27 +236,24 @@ fn write_integer(output: &mut Sink, mut n: usize) {
         let bulk = n / (4 * 0xFF);
         n %= 4 * 0xFF;
         unsafe {
-            output
-                .output
-                .get_unchecked_mut(output.pos()..output.pos() + 4 * bulk)
-                .fill(0xFF)
+            core::ptr::write_bytes(output.pos_mut_ptr(), 0xFF, 4 * bulk);
+            output.set_pos(output.pos() + 4 * bulk);
         }
-        output.set_pos(output.pos() + 4 * bulk);
     }
 
     // Handle last 1 to 4 bytes
     push_u32(output, 0xFFFFFFFF);
     // Updating output len for the remainder
-    output.set_pos(output.pos() - 4 + 1 + n / 255);
     unsafe {
+        output.set_pos(output.pos() - 4 + 1 + n / 255);
         // Write the remaining byte.
-        *output.as_mut_ptr().sub(1) = (n % 255) as u8;
+        *output.pos_mut_ptr().sub(1) = (n % 255) as u8;
     }
 }
 
 /// Handle the last bytes from the input as literals
 #[cold]
-fn handle_last_literals(output: &mut Sink, input: &[u8], start: usize) {
+fn handle_last_literals(output: &mut impl Sink, input: &[u8], start: usize) {
     let lit_len = input.len() - start;
 
     let token = token_from_literal(lit_len);
@@ -334,30 +331,28 @@ fn backtrack_match(
 /// A similar const argument could be used to disable the Prefix mode (eg. USE_PREFIX),
 /// which would impose `input_pos == 0 && input_stream_offset == 0`. Experiments didn't
 /// show significant improvement though.
-#[inline]
-pub(crate) fn compress_internal<T: HashTable, const USE_DICT: bool>(
+// Intentionally avoid inlining.
+// Empirical tests revealed it to be rarely better but often significantly detrimental.
+#[inline(never)]
+pub(crate) fn compress_internal<T: HashTable, SINK: Sink, const USE_DICT: bool>(
     input: &[u8],
     input_pos: usize,
-    output: &mut [u8],
-    output_pos: usize,
+    output: &mut SINK,
     dict: &mut T,
     ext_dict: &[u8],
     input_stream_offset: usize,
 ) -> Result<usize, CompressError> {
-    let mut output = Sink {
-        output,
-        pos: output_pos,
-    };
-    let output = &mut output;
     assert!(input_pos <= input.len());
-    assert!(ext_dict.len() <= super::WINDOW_SIZE);
-    assert!(ext_dict.len() <= input_stream_offset);
     if USE_DICT {
+        assert!(ext_dict.len() <= super::WINDOW_SIZE);
+        assert!(ext_dict.len() <= input_stream_offset);
         // Check for overflow hazard when using ext_dict
         assert!(input_stream_offset
             .checked_add(input.len())
             .and_then(|i| i.checked_add(ext_dict.len()))
             .map_or(false, |i| i <= isize::MAX as usize));
+    } else {
+        assert!(ext_dict.is_empty());
     }
     if output.capacity() - output.pos() < get_maximum_output_size(input.len() - input_pos) {
         return Err(CompressError::OutputTooSmall);
@@ -488,7 +483,7 @@ pub(crate) fn compress_internal<T: HashTable, const USE_DICT: bool>(
         //
         // The unsafe version copies blocks of 8bytes, and therefore may copy up to 7bytes more than needed.
         // This is safe, because the last 12 bytes (MF_LIMIT) are handled in handle_last_literals.
-        copy_literals_wild(output, &input, literal_start, lit_len);
+        copy_literals_wild(output, input, literal_start, lit_len);
         // write the offset in little endian.
         push_u16(output, offset);
 
@@ -503,183 +498,112 @@ pub(crate) fn compress_internal<T: HashTable, const USE_DICT: bool>(
 
 #[inline]
 #[cfg(feature = "safe-encode")]
-fn push_byte(output: &mut Sink, el: u8) {
+fn push_byte(output: &mut impl Sink, el: u8) {
     output.push(el);
 }
 
 #[inline]
 #[cfg(not(feature = "safe-encode"))]
-fn push_byte(output: &mut Sink, el: u8) {
+fn push_byte(output: &mut impl Sink, el: u8) {
     unsafe {
-        core::ptr::write(output.as_mut_ptr(), el);
-        output.pos += 1;
+        core::ptr::write(output.pos_mut_ptr(), el);
+        output.set_pos(output.pos() + 1);
     }
 }
 
 #[inline]
 #[cfg(feature = "safe-encode")]
-fn push_u16(output: &mut Sink, el: u16) {
+fn push_u16(output: &mut impl Sink, el: u16) {
     output.extend_from_slice(&el.to_le_bytes());
 }
 
 #[inline]
 #[cfg(not(feature = "safe-encode"))]
-fn push_u16(output: &mut Sink, el: u16) {
+fn push_u16(output: &mut impl Sink, el: u16) {
     unsafe {
-        output
-            .output
-            .get_unchecked_mut(output.pos()..output.pos() + 2)
-            .copy_from_slice(&el.to_le_bytes())
-    };
-    output.pos += 2;
-}
-
-#[inline]
-#[cfg(feature = "safe-encode")]
-fn push_u32(output: &mut Sink, el: u32) {
-    output.extend_from_slice(&el.to_le_bytes());
+        core::ptr::copy_nonoverlapping(el.to_le_bytes().as_ptr(), output.pos_mut_ptr(), 2);
+        output.set_pos(output.pos() + 2);
+    }
 }
 
 #[inline]
 #[cfg(not(feature = "safe-encode"))]
-fn push_u32(output: &mut Sink, el: u32) {
+fn push_u32(output: &mut impl Sink, el: u32) {
     unsafe {
-        output
-            .output
-            .get_unchecked_mut(output.pos()..output.pos() + 4)
-            .copy_from_slice(&el.to_le_bytes())
-    };
-    output.pos += 4;
+        core::ptr::copy_nonoverlapping(el.to_le_bytes().as_ptr(), output.pos_mut_ptr(), 4);
+        output.set_pos(output.pos() + 4);
+    }
 }
 
 #[inline(always)] // (always) necessary otherwise compiler fails to inline it
 #[cfg(feature = "safe-encode")]
-fn copy_literals_wild(output: &mut Sink, input: &[u8], input_start: usize, len: usize) {
+fn copy_literals_wild(output: &mut impl Sink, input: &[u8], input_start: usize, len: usize) {
     match len {
-        0..=8 => output.output[output.pos..output.pos + 8]
-            .copy_from_slice(&input[input_start..input_start + 8]),
-        9..=16 => output.output[output.pos..output.pos + 16]
-            .copy_from_slice(&input[input_start..input_start + 16]),
-        17..=24 => output.output[output.pos..output.pos + 24]
-            .copy_from_slice(&input[input_start..input_start + 24]),
-        _ => output.output[output.pos..output.pos + len]
-            .copy_from_slice(&input[input_start..input_start + len]),
+        0..=8 => output.extend_from_slice_wild(&input[input_start..input_start + 8], len),
+        9..=16 => output.extend_from_slice_wild(&input[input_start..input_start + 16], len),
+        17..=24 => output.extend_from_slice_wild(&input[input_start..input_start + 24], len),
+        _ => output.extend_from_slice_wild(&input[input_start..input_start + len], len),
     }
-    output.pos += len;
 }
 
 #[inline]
 #[cfg(not(feature = "safe-encode"))]
-fn copy_literals_wild(output: &mut Sink, input: &[u8], input_start: usize, len: usize) {
+fn copy_literals_wild(output: &mut impl Sink, input: &[u8], input_start: usize, len: usize) {
     debug_assert!(input_start + len / 8 * 8 + ((len % 8) != 0) as usize * 8 <= input.len());
     debug_assert!(output.pos() + len / 8 * 8 + ((len % 8) != 0) as usize * 8 <= output.capacity());
     unsafe {
         // Note: This used to be a wild copy loop of 8 bytes, but the compiler consistently
         // transformed it into a call to memcopy, which hurts performance significantly for
         // small copies, which are common.
+        let start_ptr = input.as_ptr().add(input_start);
         match len {
-            0..=8 => output
-                .output
-                .get_unchecked_mut(output.pos..output.pos + 8)
-                .copy_from_slice(input.get_unchecked(input_start..input_start + 8)),
-            9..=16 => output
-                .output
-                .get_unchecked_mut(output.pos..output.pos + 16)
-                .copy_from_slice(input.get_unchecked(input_start..input_start + 16)),
-            17..=24 => output
-                .output
-                .get_unchecked_mut(output.pos..output.pos + 24)
-                .copy_from_slice(input.get_unchecked(input_start..input_start + 24)),
-            _ => core::ptr::copy_nonoverlapping(
-                input.as_ptr().add(input_start),
-                output.as_mut_ptr(),
-                len,
-            ),
+            0..=8 => core::ptr::copy_nonoverlapping(start_ptr, output.pos_mut_ptr(), 8),
+            9..=16 => core::ptr::copy_nonoverlapping(start_ptr, output.pos_mut_ptr(), 16),
+            17..=24 => core::ptr::copy_nonoverlapping(start_ptr, output.pos_mut_ptr(), 24),
+            _ => core::ptr::copy_nonoverlapping(start_ptr, output.pos_mut_ptr(), len),
         }
+        output.set_pos(output.pos() + len);
     }
-    output.pos += len;
 }
 
-/// Returns the maximum output size of the compressed data.
-/// Can be used to preallocate capacity on the output vector
 #[inline]
-pub fn get_maximum_output_size(input_len: usize) -> usize {
-    16 + 4 + (input_len as f64 * 1.1) as usize
-}
-
-/// Compress all bytes of `input` into `output`.
-/// The method chooses an appropriate hashtable to lookup duplicates and calls `compress_into_with_table`.
-/// output should be preallocated with a size of `get_maximum_output_size`.
-///
-/// Returns the number of bytes written (compressed) into `output`.
-#[inline]
-pub fn compress_into(
+pub(crate) fn compress_into_sink(
     input: &[u8],
-    output: &mut [u8],
-    output_pos: usize,
+    output: &mut impl Sink,
 ) -> Result<usize, CompressError> {
     let (dict_size, dict_bitshift) = get_table_size(input.len());
     if input.len() < u16::MAX as usize {
         let mut dict = HashTableU16::new(dict_size, dict_bitshift);
-        compress_internal::<_, false>(input, 0, output, output_pos, &mut dict, b"", 0)
+        compress_internal::<_, _, false>(input, 0, output, &mut dict, b"", 0)
     } else if input.len() < u32::MAX as usize {
         let mut dict = HashTableU32::new(dict_size, dict_bitshift);
-        compress_internal::<_, false>(input, 0, output, output_pos, &mut dict, b"", 0)
+        compress_internal::<_, _, false>(input, 0, output, &mut dict, b"", 0)
     } else {
         let mut dict = HashTableUsize::new(dict_size, dict_bitshift);
-        compress_internal::<_, false>(input, 0, output, output_pos, &mut dict, b"", 0)
+        compress_internal::<_, _, false>(input, 0, output, &mut dict, b"", 0)
     }
 }
 
-/// Compress all bytes of `input` into `output`.
-/// The method chooses an appropriate hashtable to lookup duplicates and calls `compress_into_with_table`.
-/// output should be preallocated with a size of `get_maximum_output_size`.
-///
-/// Returns the number of bytes written (compressed) into `output`.
+/// Same as compress_into_sink but with supports external dictionary
 #[inline]
-pub fn compress_into_with_dict(
+pub(crate) fn compress_into_sink_with_dict(
     input: &[u8],
-    output: &mut [u8],
-    output_pos: usize,
+    output: &mut impl Sink,
     mut dict_data: &[u8],
 ) -> Result<usize, CompressError> {
     let (dict_size, dict_bitshift) = get_table_size(input.len());
     if dict_data.len() + input.len() < u16::MAX as usize {
         let mut dict = HashTableU16::new(dict_size, dict_bitshift);
         init_dict(&mut dict, &mut dict_data);
-        compress_internal::<_, true>(
-            input,
-            0,
-            output,
-            output_pos,
-            &mut dict,
-            dict_data,
-            dict_data.len(),
-        )
+        compress_internal::<_, _, true>(input, 0, output, &mut dict, dict_data, dict_data.len())
     } else if dict_data.len() + input.len() < u32::MAX as usize {
         let mut dict = HashTableU32::new(dict_size, dict_bitshift);
         init_dict(&mut dict, &mut dict_data);
-        compress_internal::<_, true>(
-            input,
-            0,
-            output,
-            output_pos,
-            &mut dict,
-            dict_data,
-            dict_data.len(),
-        )
+        compress_internal::<_, _, true>(input, 0, output, &mut dict, dict_data, dict_data.len())
     } else {
         let mut dict = HashTableUsize::new(dict_size, dict_bitshift);
         init_dict(&mut dict, &mut dict_data);
-        compress_internal::<_, true>(
-            input,
-            0,
-            output,
-            output_pos,
-            &mut dict,
-            dict_data,
-            dict_data.len(),
-        )
+        compress_internal::<_, _, true>(input, 0, output, &mut dict, dict_data, dict_data.len())
     }
 }
 
@@ -697,30 +621,49 @@ fn init_dict<T: HashTable>(dict: &mut T, dict_data: &mut &[u8]) {
     }
 }
 
+/// Returns the maximum output size of the compressed data.
+/// Can be used to preallocate capacity on the output vector
 #[inline]
-pub(crate) fn get_output_vec(input_len: usize) -> Vec<u8> {
-    let max_size = get_maximum_output_size(input_len);
+pub fn get_maximum_output_size(input_len: usize) -> usize {
+    16 + 4 + (input_len as f64 * 1.1) as usize
+}
 
-    let mut compressed = Vec::with_capacity(max_size);
-    #[cfg(not(feature = "safe-encode"))]
-    {
-        unsafe { compressed.set_len(max_size) };
-    }
-    #[cfg(feature = "safe-encode")]
-    {
-        compressed.resize(max_size, 0);
-    }
-    compressed
+/// Compress all bytes of `input` into `output`.
+/// The method chooses an appropriate hashtable to lookup duplicates and calls `compress_into_with_table`.
+/// output should be preallocated with a size of `get_maximum_output_size`.
+///
+/// Returns the number of bytes written (compressed) into `output`.
+#[inline]
+pub fn compress_into(input: &[u8], output: &mut [u8]) -> Result<usize, CompressError> {
+    compress_into_sink(input, &mut SliceSink::new(output, 0))
+}
+
+/// Compress all bytes of `input` into `output`.
+/// The method chooses an appropriate hashtable to lookup duplicates and calls `compress_into_with_table`.
+/// output should be preallocated with a size of `get_maximum_output_size`.
+///
+/// Returns the number of bytes written (compressed) into `output`.
+#[inline]
+pub fn compress_into_with_dict(
+    input: &[u8],
+    output: &mut [u8],
+    dict_data: &[u8],
+) -> Result<usize, CompressError> {
+    compress_into_sink_with_dict(input, &mut SliceSink::new(output, 0), dict_data)
 }
 
 /// Compress all bytes of `input` into `output`. The uncompressed size will be prepended as a little endian u32.
 /// Can be used in conjunction with `decompress_size_prepended`
 #[inline]
 pub fn compress_prepend_size(input: &[u8]) -> Vec<u8> {
-    let mut compressed = get_output_vec(4 + input.len());
-    let mut sink: Sink = (&mut compressed).into();
-    push_u32(&mut sink, input.len() as u32);
-    let compressed_len = compress_into(input, &mut sink.output, sink.pos).unwrap();
+    let max_compressed_size = get_maximum_output_size(input.len());
+    let mut compressed = Vec::with_capacity(4 + max_compressed_size);
+    compressed.extend_from_slice(&(input.len() as u32).to_le_bytes());
+    let compressed_len = compress_into_sink(
+        input,
+        &mut vec_sink_for_compression(&mut compressed, 4, 0, max_compressed_size),
+    )
+    .unwrap();
     compressed.truncate(4 + compressed_len);
     compressed
 }
@@ -728,9 +671,13 @@ pub fn compress_prepend_size(input: &[u8]) -> Vec<u8> {
 /// Compress all bytes of `input`.
 #[inline]
 pub fn compress(input: &[u8]) -> Vec<u8> {
-    let mut compressed = get_output_vec(input.len());
-    let mut sink: Sink = (&mut compressed).into();
-    let compressed_len = compress_into(input, &mut sink.output, sink.pos).unwrap();
+    let max_compressed_size = get_maximum_output_size(input.len());
+    let mut compressed = Vec::with_capacity(max_compressed_size);
+    let compressed_len = compress_into_sink(
+        input,
+        &mut vec_sink_for_compression(&mut compressed, 0, 0, max_compressed_size),
+    )
+    .unwrap();
     compressed.truncate(compressed_len);
     compressed
 }
@@ -738,8 +685,14 @@ pub fn compress(input: &[u8]) -> Vec<u8> {
 /// Compress all bytes of `input` with an external dictionary.
 #[inline]
 pub fn compress_with_dict(input: &[u8], ext_dict: &[u8]) -> Vec<u8> {
-    let mut compressed = get_output_vec(input.len());
-    let compressed_len = compress_into_with_dict(input, &mut compressed, 0, ext_dict).unwrap();
+    let max_compressed_size = get_maximum_output_size(input.len());
+    let mut compressed = Vec::with_capacity(max_compressed_size);
+    let compressed_len = compress_into_sink_with_dict(
+        input,
+        &mut vec_sink_for_compression(&mut compressed, 0, 0, max_compressed_size),
+        ext_dict,
+    )
+    .unwrap();
     compressed.truncate(compressed_len);
     compressed
 }
@@ -748,12 +701,15 @@ pub fn compress_with_dict(input: &[u8], ext_dict: &[u8]) -> Vec<u8> {
 /// Can be used in conjunction with `decompress_size_prepended_with_dict`
 #[inline]
 pub fn compress_prepend_size_with_dict(input: &[u8], ext_dict: &[u8]) -> Vec<u8> {
-    // In most cases, the compression won't expand the size, so we set the input size as capacity.
-    let mut compressed = get_output_vec(4 + input.len());
-    let mut sink: Sink = (&mut compressed).into();
-    push_u32(&mut sink, input.len() as u32);
-    let compressed_len =
-        compress_into_with_dict(input, &mut sink.output, sink.pos, ext_dict).unwrap();
+    let max_compressed_size = get_maximum_output_size(input.len());
+    let mut compressed = Vec::with_capacity(4 + max_compressed_size);
+    compressed.extend_from_slice(&(input.len() as u32).to_le_bytes());
+    let compressed_len = compress_into_sink_with_dict(
+        input,
+        &mut vec_sink_for_compression(&mut compressed, 4, 0, max_compressed_size),
+        ext_dict,
+    )
+    .unwrap();
     compressed.truncate(4 + compressed_len);
     compressed
 }
@@ -878,7 +834,7 @@ mod tests {
         let input: &[u8] = &[
             10, 12, 14, 16, 18, 10, 12, 14, 16, 18, 10, 12, 14, 16, 18, 10, 12, 14, 16, 18,
         ];
-        let _out = compress(&input);
+        let _out = compress(input);
     }
 
     #[test]
@@ -887,7 +843,7 @@ mod tests {
             10, 12, 14, 16, 18, 10, 12, 14, 16, 18, 10, 12, 14, 16, 18, 10, 12, 14, 16, 18,
         ];
         let dict = input;
-        let compressed = compress_with_dict(&input, &dict);
+        let compressed = compress_with_dict(input, dict);
         assert_lt!(compressed.len(), compress(input).len());
 
         assert!(compressed.len() < compress(input).len());
@@ -895,8 +851,7 @@ mod tests {
         let uncomp_size = crate::block::decompress::decompress_into_with_dict(
             &compressed,
             &mut uncompressed,
-            0,
-            &dict,
+            dict,
         )
         .unwrap();
         uncompressed.truncate(uncomp_size);
@@ -909,7 +864,7 @@ mod tests {
             10, 12, 14, 16, 18, 10, 12, 14, 16, 18, 10, 12, 14, 16, 18, 10, 12, 14, 16, 18,
         ];
         let dict = input;
-        let compressed = compress_with_dict(&input, &dict);
+        let compressed = compress_with_dict(input, dict);
         assert_lt!(compressed.len(), compress(input).len());
 
         let mut uncompressed = vec![0u8; input.len() * 2];
@@ -917,14 +872,15 @@ mod tests {
         let dict_cutoff = dict.len() / 2;
         let output_start = dict.len() - dict_cutoff;
         uncompressed[..output_start].copy_from_slice(&dict[dict_cutoff..]);
-
-        let uncomp_len = crate::block::decompress::decompress_into_with_dict(
-            &compressed,
-            &mut uncompressed,
-            output_start,
-            &dict[..dict_cutoff],
-        )
-        .unwrap();
+        let uncomp_len = {
+            let mut sink = SliceSink::new(&mut uncompressed[..], output_start);
+            crate::block::decompress::decompress_internal::<_, true>(
+                &compressed,
+                &mut sink,
+                &dict[..dict_cutoff],
+            )
+            .unwrap()
+        };
         assert_eq!(input.len(), uncomp_len);
         assert_eq!(
             input,
@@ -955,18 +911,18 @@ mod tests {
         assert_le!(out.len(), 15);
 
         // dict uncompressible
-        let out = compress_with_dict(&aaas[..11], &aaas);
+        let out = compress_with_dict(&aaas[..11], aaas);
         assert_gt!(out.len(), 11);
         // compressible
-        let out = compress_with_dict(&aaas[..12], &aaas);
+        let out = compress_with_dict(&aaas[..12], aaas);
         // According to the spec this _could_ compres, but it doesn't in this lib
         // as it aborts compression for any input len < LZ4_MIN_LENGTH
         assert_gt!(out.len(), 12);
-        let out = compress_with_dict(&aaas[..13], &aaas);
+        let out = compress_with_dict(&aaas[..13], aaas);
         assert_le!(out.len(), 13);
-        let out = compress_with_dict(&aaas[..14], &aaas);
+        let out = compress_with_dict(&aaas[..14], aaas);
         assert_le!(out.len(), 14);
-        let out = compress_with_dict(&aaas[..15], &aaas);
+        let out = compress_with_dict(&aaas[..15], aaas);
         assert_le!(out.len(), 15);
     }
 
@@ -974,7 +930,7 @@ mod tests {
     fn test_dict_size() {
         let dict = vec![b'a'; 1024 * 1024];
         let input = &b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaa"[..];
-        let compressed = compress_prepend_size_with_dict(&input, &dict);
+        let compressed = compress_prepend_size_with_dict(input, &dict);
         let decompressed =
             crate::block::decompress_size_prepended_with_dict(&compressed, &dict).unwrap();
         assert_eq!(decompressed, input);
